// Initial version of FormItSectorMenu retrieved in October 2015 from
// https://doc.qt.io/archives/qq/qq11-piemenu.html

// FormItPieMenu is based on QtPieMenu with the following copyright
/****************************************************************************
**
** Copyright (c) 2013 Digia Plc and/or its subsidiary(-ies).
** Contact: http://www.qt-project.org/legal
**
** This file is part of a Qt Solutions component.
**
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Solutions Commercial License Agreement provided
** with the Software or, alternatively, in accordance with the terms
** contained in a written agreement between you and Digia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Digia gives you certain
** additional rights. These rights are described in the Digia Qt LGPL
** Exception version 1.1, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** Please note Third Party Software included with Qt Solutions may impose
** additional restrictions and it is the user's responsibility to ensure
** that they have met the licensing requirements of the GPL, LGPL, or Qt
** Solutions Commercial license and the relevant license of the Third
** Party Software they are using.
**
** If you are unsure which license is appropriate for your use, please
** contact Digia at http://www.qt-project.org/legal.
**
****************************************************************************/

#include "FormItPieMenu.h"

#include <QAction>
#include <QBitmap>
#include <QHelpEvent>
#include <QPainter>
#include <QToolTip>
#include <qtpieaction.h>

#define CENTER_IS_CLEAR

const static double PI = 3.14159265358979323846264338327950288419717;
const static double TWOPI = 2.0*PI;
const QColor kIconHighlightColor = QColor::fromRgb(195, 232, 244, 0);


FormItPieMenu::FormItPieMenu(QWidget* parent, uint innerRad, uint outerRad):
    QtPieMenu(QString{}, parent, nullptr, innerRad, outerRad)
{
#ifdef CENTER_IS_CLEAR
    setAttribute(Qt::WA_NativeWindow);
    setAttribute(Qt::WA_TranslucentBackground);
    setWindowFlags(Qt::Popup | Qt::FramelessWindowHint | Qt::NoDropShadowWindowHint);
#endif
    
}

void FormItPieMenu::insertItem(QAction *pAction, int index)
{
    // Out of bounds indexes are treated as appends.
    if (index < 0 || index > int(count()))
        index = count();

    QtPieMenu::insertItem(pAction, index);

    if (nullptr == pAction)
    {
        setItemEnabled(false, index);
    }
}

void FormItPieMenu::insertItem(QIcon icon, FormItSectorMenu *item, int index)
{
    // Out of bounds indexes are treated as appends.
    if (index < 0 || index > int(count()))
        index = count();

    QtPieMenu::insertItem(item, index);

    setItemIcon(icon, index);
}

// This function is only needed until we can have a translucent center.
void FormItPieMenu::generateMask(QBitmap *
    /*mask*/)
{
#ifndef CENTER_IS_CLEAR
    QPainter p(mask);
    p.setPen(Qt::color1);
    p.setBrush(Qt::color1);
    p.drawEllipse(mask->rect().adjusted(0, 0, -1, -1));
    int offset = outerRadius() - innerRadius();
    p.setPen(Qt::color0);
    p.setBrush(Qt::color0);
    p.drawEllipse(mask->rect().adjusted(offset, offset, -offset - 1, -offset - 1));
#endif
}

inline QColor translucent(QColor color)
{
    color.setAlpha(172);
    return color;
}

inline QBrush translucent(QBrush brush)
{
    brush.setColor(translucent(brush.color()));
    return brush;
}

inline double DegreesToRadians(double dAngle)
{
    return dAngle * PI / 180.0;
}

inline double RadiansToDegrees(double dAngle)
{
    return dAngle * 180.0 / PI;
}

inline int DegreesToPainterAngle(double dAngle)
{
    return static_cast<int>(dAngle * 16);
}

inline double PainterAngleToDegrees(int nAngle)
{
    return nAngle / 16;
}

inline int RadiansToPainterAngle(double dAngle)
{
    return DegreesToPainterAngle(RadiansToDegrees(dAngle));
}

inline double PainterAngleToRadians(int nAngle)
{
    return DegreesToRadians(PainterAngleToDegrees(nAngle));
}

// Draws a single segment of the menu ring.
static void DrawSegment(QPainter &p, int nThickness, QRect rect,
                        double dStartAngle, double dDeltaAngle)
{
    // Convert the angles from radians.
    dStartAngle = RadiansToDegrees(dStartAngle);
    dDeltaAngle = RadiansToDegrees(dDeltaAngle);

    // Move to the start position of the circular wedge
    QPainterPath path;
    path.arcMoveTo(rect, dStartAngle);

    // Remember this position for future use.
    auto pt1 = path.currentPosition();

    // Create the outer arc
    path.arcTo(rect, dStartAngle, dDeltaAngle);

    // Adjust the rect to be the inside arc
    rect -= QMargins{nThickness, nThickness, nThickness, nThickness};

    // Create a temporary path to get the endpoint of the line.
    QPainterPath temp;
    temp.arcMoveTo(rect, dStartAngle + dDeltaAngle);

    // Create the line connecting to the inner arc
    path.lineTo(temp.currentPosition());

    // Create the inner arc
    path.arcTo(rect, dStartAngle + dDeltaAngle, -dDeltaAngle);

    // Create the line back to the outer arc.
    path.lineTo(pt1);

    // Finally draw the path
    p.drawPath(path);
}

// Draw the pie menu and blit it onto the screen.
void FormItPieMenu::paintEvent(QPaintEvent *)
{
    double dOuterRadius  = outerRadius();
    double dInnerRadius  = innerRadius();
    double dMiddleRadius = (dInnerRadius + dOuterRadius) / 2.0;
    double dThickness    = dOuterRadius - dInnerRadius;
            
    auto &rPalette = palette();

#ifndef CENTER_IS_CLEAR
    auto brushBackground    = rPalette.background();
    auto brushHighlight     = rPalette.highlight();
#else
    auto brushBackground    = translucent(QColor::fromRgb(255, 255, 255));
    auto brushHighlight     = QColor::fromRgb(195, 232, 244, 250);
#endif
    auto brushTextRegular   = rPalette.text();
    auto brushTextHighlight = rPalette.highlightedText();


    pix.fill(Qt::transparent);
    
    QPainter p(&pix);
    p.setRenderHint(QPainter::Antialiasing, true);
    QRect adjustedRect = pix.rect().adjusted(0, 0, -1, -1);
        
    // Get the center of the menu.
    auto ptMenuCenter = adjustedRect.center();
    double xMenuCenter = ptMenuCenter.x(), yMenuCenter = ptMenuCenter.y();

#ifndef CENTER_IS_CLEAR
    bool bCancelIsActive = ((highlightedItem() == -1) ||
                            !isItemEnabled(highlightedItem()));

    // Fill the center and draw the X.
    p.setBrush(bCancelIsActive ? brushHighlight : brushBackground);
    p.setPen(Qt::NoPen);

    QRect innerRect(xMenuCenter - dInnerRadius, yMenuCenter - dInnerRadius,
                    dInnerRadius * 2 + 1, dInnerRadius * 2 + 1);
    p.drawPie(innerRect, 0, DegreesToPainterAngle(360.0));

    if (bCancelIsActive)
        p.setPen({brushTextHighlight.color(), 2});
    else
        p.setPen({brushTextRegular.color(), 2});

    int nOffset = innerRadius() / 3;
    innerRect -= QMargins(nOffset, nOffset, nOffset, nOffset);
    p.drawLine(innerRect.topLeft(),  innerRect.bottomRight());
    p.drawLine(innerRect.topRight(), innerRect.bottomLeft());
#endif

    // Don't draw item borders.
    p.setPen(Qt::NoPen);

    // Draw little arrows on the edge of the items that have a submenu.
    // Smooth edges of the arrow by drawing with a pen that has the average
    // color of the foreground and the background.
    auto averagePen = [](QBrush const &b1, QBrush const &b2) -> QPen
    {
        auto c1 = b1.color(), c2 = b2.color();
        return QColor{(c1.red()   + c2.red())   / 2,
                      (c1.green() + c2.green()) / 2,
                      (c1.blue()  + c2.blue())  / 2};
    };
    auto penArrowRegular   = averagePen(brushTextRegular,   brushBackground);
    auto penArrowHighlight = averagePen(brushTextHighlight, brushHighlight);

    for (int i = 0, nItems = count(); i < nItems; ++i)
    {
        // Find the angles of the edges of the menu item.
        double angle1 = indexToAngle(i,     nItems, true);
        double angle2 = indexToAngle(i + 1, nItems, true);

        if (angle2 < angle1)
            angle2 += TWOPI;

        // Find the size and the angle of the center of the menu item.
        double size = angle2 - angle1;
        double angle = (angle1 + angle2) / 2.0;

        // Is this menu item highlighted.
        bool bHL = ((i == highlightedItem() && isItemEnabled(i)));

        // Draw the item background.
        p.setBrush(bHL ? brushHighlight : brushBackground);
        DrawSegment(p, dThickness, adjustedRect, angle1, size);

        // Get the item image
        auto icon = itemIcon(i);

        if (icon.isNull())
            continue;

        // Get the sin and cosine of the item center angle angle.
        double cosCenter = cos(angle), sinCenter = sin(angle);

        // Get a pixmap for the image.
		auto dpiScale = logicalDpiX() / 96.0f;
		int sz = static_cast<int>(32 * dpiScale);
        QPixmap item_pix = icon.pixmap({sz, sz}, QIcon::Normal, QIcon::On);
        
        int xImageCenter = xMenuCenter + static_cast<int>(dMiddleRadius * cosCenter);
        int yImageCenter = yMenuCenter - static_cast<int>(dMiddleRadius * sinCenter);
        int imageWidth  = item_pix.width();
        int imageHeight = item_pix.height();

        // Draw the image.
        p.drawPixmap(xImageCenter - imageWidth  / 2,
            yImageCenter - imageHeight / 2,
            imageWidth, imageHeight, item_pix);

#ifdef CENTER_IS_CLEAR
        if(bHL)
        {
            // This is a goofy hack to "colorize" the icons. Probably take this out
            // when we bring in the highlighted images, or decide we don't want highlighted images.
            auto img = item_pix.toImage();
            auto alphaMaskMap = QPixmap::fromImage(img.createAlphaMask());
            QRegion region(alphaMaskMap);

            region.translate(
                xImageCenter - imageWidth  / 2,
                yImageCenter - imageHeight / 2);

            p.setClipRegion(region);
            
            p.fillRect(xImageCenter - imageWidth  / 2,
                yImageCenter - imageHeight / 2,
                imageWidth, imageHeight, kIconHighlightColor);
            
            p.setClipping(false);
        }
#endif

        // Look to see if we need to draw a menu arrow.
        if (itemType(i) != SubMenu)
            continue;

        // Distance from edge of pie to tip of arrow.
        const double h1 = 2;

        // Height of arrow.
        const double h2 = 6;

        // Width of arrow in radians.
        const double w = 0.1;

        const double toprad   = dOuterRadius - h1;
        const double leftrad  = dOuterRadius - h2;
        const double rightrad = dOuterRadius - h2;
        const int topx   = static_cast<int>( toprad   * cosCenter);
        const int topy   = static_cast<int>(-toprad   * sinCenter);
        const int leftx  = static_cast<int>( leftrad  * cos(angle + w));
        const int lefty  = static_cast<int>(-leftrad  * sin(angle + w));
        const int rightx = static_cast<int>( rightrad * cos(angle - w));
        const int righty = static_cast<int>(-rightrad * sin(angle - w));

        QPolygon ar(3);
        ar.setPoint(0, xMenuCenter + topx,   yMenuCenter + topy);
        ar.setPoint(1, xMenuCenter + rightx, yMenuCenter + righty);
        ar.setPoint(2, xMenuCenter + leftx,  yMenuCenter + lefty);

        p.setBrush(bHL ? brushTextHighlight : brushTextRegular);
        p.setPen(bHL ? penArrowHighlight : penArrowRegular);
        p.drawPolygon(ar);
        p.setPen(Qt::NoPen);
    }

    QPainter p2(this);
    p2.drawPixmap(QPoint{}, pix);
}

bool FormItPieMenu::event(QEvent *event)
{
    if (event->type() != QEvent::ToolTip)
        return QtPieMenu::event(event);

    auto helpEvent = dynamic_cast<QHelpEvent *>(event);

    if (nullptr == helpEvent)
        return true;

    QToolTip::showText(helpEvent->globalPos(), itemTooltip(highlightedItem()),
                       this, QRect(), toolTipDuration());
    
    return true;
}

FormItSectorMenu::FormItSectorMenu(QWidget *parent, uint innerRad,
                                   uint outerRad) :
    FormItPieMenu(parent, innerRad, outerRad)
{
    int nThickness = outerRad - innerRad;
    innerRad += menuDepth() * nThickness;
    setInnerRadius(innerRad);
    setOuterRadius(innerRad + nThickness);

    setFixedSize(2 * outerRadius() + 1, 2 * outerRadius() + 1);
}

int FormItSectorMenu::menuDepth() const
{
    const QObject *pie = this;
    int depth = 0;

    while (pie->parent() && pie->parent()->inherits("QtPieMenu"))
    {
        pie = pie->parent();
        ++depth;
    }

    return depth;
}

void FormItSectorMenu::generateMask(QBitmap *mask)
{
    if (menuDepth() == 0)
    {
        m_startAngle = DegreesToPainterAngle(90);
        m_arcLength = DegreesToPainterAngle(360);
        m_dStartAngleRad = DegreesToRadians(90);
        m_dArcLengthRad = DegreesToRadians(360);
    }
    else if(auto parentPie = dynamic_cast<FormItPieMenu *>(parent()))
    {
        for (int i = 0, n = parentPie->count(); i < n;  ++i)
        {
            if (parentPie->subMenuAt(i) == this)
            {
                if(auto parentSector = dynamic_cast<FormItSectorMenu *>(parentPie))
                    m_arcLength = parentSector->m_arcLength / parentPie->count();
                else
                    m_arcLength = DegreesToPainterAngle(360) / parentPie->count();
                m_startAngle = DegreesToPainterAngle(90) + m_arcLength * (i - 1) / 2;
                m_dStartAngleRad = PainterAngleToRadians(m_startAngle);
                m_dArcLengthRad = PainterAngleToRadians(m_arcLength);
                break;
            }
        }
    }

    QPainter painter(mask);
    painter.setPen(Qt::color1);
    painter.setBrush(Qt::color1);
    painter.drawPie(0, 0, outerRadius() * 2, outerRadius() * 2,
                    m_startAngle, m_arcLength);

    if (innerRadius() > 0)
    {
        QPoint center = rect().center();
        painter.setPen(Qt::color0);
        painter.setBrush(Qt::color0);
        painter.drawPie(center.x() - innerRadius(),
                        center.y() - innerRadius(),
                        innerRadius() * 2, innerRadius() * 2, 0,
                        DegreesToPainterAngle(360));
    }
}

int FormItSectorMenu::indexAt(const QPoint &pos)
{
    if (count() == 0)
        return -1;

    int sliceSize = m_arcLength / count();

    if (sliceSize == 0)
        return -1;

    int angle = RadiansToPainterAngle(angleAt(pos));

    if (angle - m_startAngle < 0)
        angle += DegreesToPainterAngle(360);

    angle -= m_startAngle;

    int sector = angle / sliceSize;

    if (sector < 0 || sector >= count())
        return -1;

    return sector;
}

void FormItSectorMenu::reposition()
{
    if (auto parentPie = dynamic_cast<FormItPieMenu *>(parent()))
    {
        QPoint center = parentPie->geometry().center();
        move(center.x() - outerRadius(), center.y() - outerRadius());
    }
}

void FormItSectorMenu::paintEvent(QPaintEvent *)
{
    QPainter painter(this);
    painter.setPen(Qt::NoPen);

    auto &rPalette = palette();
    auto brushBackground = rPalette.background();
    auto brushHighlight = rPalette.highlight();

    // Get the center of the menu.
    auto rectAdjusted = rect().adjusted(0, 0, -1, -1);
    auto ptMenuCenter = rectAdjusted.center();
    double xMenuCenter = ptMenuCenter.x(), yMenuCenter = ptMenuCenter.y();

    int nItems = count();
    double dOuterRadius  = outerRadius();
    double dInnerRadius  = innerRadius();
    double dMiddleRadius = (dInnerRadius + dOuterRadius) / 2.0;
    double dThickness    = dOuterRadius - dInnerRadius;
    double dSliceRad     = m_dArcLengthRad / nItems;

    for (int i = 0; i < nItems; ++i)
    {
        if (i == highlightedItem())
            painter.setBrush(brushHighlight);
        else
            painter.setBrush(brushBackground);

        double dSegmentStartAngle = m_dStartAngleRad + dSliceRad * i;
        DrawSegment(painter, dThickness, rectAdjusted,
                    dSegmentStartAngle, dSliceRad);

        double dImageCenterAngle = dSegmentStartAngle + dSliceRad / 2.0;
        int x = static_cast<int>(cos(dImageCenterAngle) * dMiddleRadius);
        int y = static_cast<int>(sin(dImageCenterAngle) * dMiddleRadius);

        QPixmap item_pix = itemIcon(i).pixmap({32, 32}, QIcon::Normal, QIcon::On);

        int imageWidth = item_pix.width();
        int imageHeight = item_pix.height();
        painter.drawPixmap(xMenuCenter + x - imageWidth  / 2,
                           yMenuCenter - y - imageHeight / 2,
                           imageWidth, imageHeight, item_pix);
    }
}
